#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
from typing import Tuple
import os
import re
import json
import io

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary import CkanMap
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.policies.data_format_policy import (SingleValueListPolicy, ValueListPolicy, StringValueSpecification,
                                                            ListChoiceMode, CustomFieldsPolicy, CustomFieldSpecification,
                                                            GroupedValueListPolicy, ErrorLevel, DataPolicyError,
                                                            StringMatchMode, TagListPolicy, TagGroupsListPolicy)

from ckanapi_harvesters.builder.specific.configuration_builder import ConfigurationBuilder
from ckanapi_harvesters.builder.example import example_package_xls
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))


enable_write = False  # be careful, setting this to True could erase your CKAN data format policy !!!


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(False)
    ckan.verbose_policy = True

    policy = CkanPackageDataFormatPolicy()
    policy.resource_format = SingleValueListPolicy(ValueListPolicy([StringValueSpecification("CSV")]), extra_values=ListChoiceMode.NoExtra)
    policy.package_custom_fields = CustomFieldsPolicy([
                                                        CustomFieldSpecification(key="New field", mandatory=True),
                                                        CustomFieldSpecification(key="Algorithm", values=["Random"], match_mode=StringMatchMode.Match, mandatory=True)], error_level=ErrorLevel.Error)
    policy.package_tags = TagGroupsListPolicy([TagListPolicy([StringValueSpecification("Test")], group_name="Vocabulary_Test")])
    policy.package_mandatory_attributes = {"description", "author"}
    ckan.policy = policy

    policy_dict = policy.to_dict(sets_as_lists=False)
    policy_from_dict = CkanPackageDataFormatPolicy.from_dict(policy_dict)
    policy_dict_bis = policy_from_dict.to_dict(sets_as_lists=False)
    assert(policy_dict == policy_dict_bis)

    # serialisation
    policy_dict = policy.to_dict(sets_as_lists=True)
    policy_json_file = os.path.abspath("policy_py.json")
    with open(policy_json_file, "w") as f:
        json.dump(policy_dict, f, indent=4)
    with open(policy_json_file, "r") as f:
        policy_json_dict = json.load(f)
    # policy_json = CkanPackageDataFormatPolicy.from_dict(policy_json_dict)
    policy_json = CkanPackageDataFormatPolicy.from_json(policy_json_file)
    assert(policy_dict == policy_json_dict)

    # test on local definition (offline)
    buffer = {}
    success = mdl.local_policy_check(policy, buffer=buffer)
    # test if an error is raised (this mode does not display all messages)
    try:
        success = mdl.local_policy_check(policy)
    except DataPolicyError as e:
        print("Exception: " + str(e))
        assert(not success)
    else:
        print(f"No exception / success={success}")

    # test on remote definition (CKAN API)
    print("Test on remote")
    buffer = {}
    ckan.map_resources(mdl.package_name)
    success = ckan.policy_check(buffer=buffer)

    ckan.set_verbosity(True)

    # update vocabularies: deprecated
    # if enable_write:
    #     ckan.vocabularies_clear()
    #     ckan.initiate_vocabularies_from_policy(policy, remove_others=True)

    # upload default policy
    config_ckan = ConfigurationBuilder(ckan, ckan.owner_org)
    if enable_write:
        config_ckan.patch_policy(ckan, policy, reduced_size=False)

    # download default policy
    # default_policy = ckan.load_default_policy(force=True)
    config_ckan.load_default_policy(ckan)

    # check all packages against policy
    ckan.owner_org = None
    print(" ")
    buffer = {}
    config_ckan.policy_check(ckan, policy=policy, buffer=buffer, verbose=True)

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements the field definition
"""
from typing import Union
import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_model import CkanFieldType, CkanField
from ckanapi_harvesters.auxiliary.ckan_auxiliary import CkanFieldInternalAttrs
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, _bool_from_string




class BuilderField:
    def __init__(self, *, name:str=None, type_override:CkanFieldType=None,
                 label:str=None, description:str=None):
        self.name: str = name
        self.type_override: Union[CkanFieldType,None] = type_override
        self.label: Union[str,None] = label
        self.description: Union[str,None] = description
        self.is_index: Union[bool,None] = None
        self.uniquekey: Union[bool,None] = None
        self.notnull: Union[bool,None] = None
        self.options_string: Union[str,None] = None
        self.internal_attrs: CkanFieldInternalAttrs = CkanFieldInternalAttrs()
        self.comment: Union[str,None] = None

    def __str__(self):
        return f"Field builder for {self.name}"

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderField()
        dest.name = self.name
        dest.type_override = self.type_override
        dest.label = self.label
        dest.description = self.description
        dest.is_index = self.is_index
        dest.uniquekey = self.uniquekey
        dest.notnull = self.notnull
        dest.options_string = self.options_string
        dest.internal_attrs = self.internal_attrs.copy()
        dest.comment = self.comment
        return dest

    def _load_from_df_row(self, row: pd.Series):
        self.name = _string_from_element(row["field name"]).strip()
        type_override_string = _string_from_element(row["type override"])
        self.type_override = None
        if type_override_string is not None:
            self.type_override = CkanFieldType.from_str(type_override_string)
        else:
            self.type_override = None
        self.label = None
        self.description = None
        if "label" in row.keys():
            self.label = _string_from_element(row["label"])
        if "description" in row.keys():
            self.description = _string_from_element(row["description"])
        if "index" in row.keys():
            self.is_index = _bool_from_string(row["index"], default_value=None)
        if "unique" in row.keys():
            self.uniquekey = _bool_from_string(row["unique"], default_value=None)
        if "not null" in row.keys():
            self.notnull = _bool_from_string(row["not null"], default_value=None)
        if "options" in row.keys():
            self.options_string = _string_from_element(row["options"])
        if "comment" in row.keys():
            self.comment = _string_from_element(row["comment"])
        self.internal_attrs.init_from_native_type(self.type_override)
        self.internal_attrs.init_from_options_string(self.options_string)

    @staticmethod
    def from_df_row(row: pd.Series) -> "BuilderField":
        field_builder = BuilderField()
        field_builder._load_from_df_row(row)
        return field_builder

    def _to_dict(self) -> dict:
        return {
            "Field Name": self.name,
            "Type override": str(self.type_override) if self.type_override is not None else "",
            "Label": self.label if self.label else "",
            "Description": self.description if self.description else "",
            "Index": str(self.is_index) if self.is_index is not None else "",
            "Unique": str(self.uniquekey) if self.uniquekey is not None else "",
            "Not null": str(self.notnull) if self.notnull is not None else "",
            "Options": self.options_string if self.options_string else "",
            "Comment": self.comment if self.comment else "",
        }

    def _to_ckan_field(self) -> CkanField:
        field_info = CkanField(name=self.name, data_type=str(self.type_override) if self.type_override is not None else "",
                               notes=self.description, label=self.label)
        field_info.is_index = self.is_index
        field_info.uniquekey = self.uniquekey
        field_info.notnull = self.notnull
        field_info.internal_attrs = self.internal_attrs.copy()
        return field_info

    def _to_ckan_dict(self) -> dict:
        return self._to_ckan_field().to_ckan_dict()

    @staticmethod
    def _from_ckan_field(field_info: CkanField) -> "BuilderField":
        field_row = pd.Series({"field name": field_info.name,
                               "type override": str(field_info.data_type),  # if field_info.type_override else "",
                               "label": field_info.label,
                               "description": field_info.notes,
                               "index": field_info.is_index,
                               "unique": field_info.uniquekey,
                               "not null": field_info.notnull,
                               })
        field_builder = BuilderField()
        field_builder._load_from_df_row(field_row)
        field_builder.internal_attrs = field_info.internal_attrs.copy()
        return field_builder




#!python3
# -*- coding: utf-8 -*-
"""
CKAN error types
"""
from typing import Iterable
import requests

# import to make these error codes available from here:
from ckanapi_harvesters.auxiliary.ckan_action import (CkanActionError, CkanAuthorizationError, CkanNotFoundError,
                                                      CkanSqlCapabilityError)
from ckanapi_harvesters.auxiliary.path import BaseDirUndefError


## Specific error types ------------------
class ApiKeyFileError(Exception):
    pass

class InvalidParameterError(Exception):
    pass

class FileOrDirNotExistError(Exception):
    def __init__(self, path: str):
        super().__init__(f"Path doesn't lead to a file or directory: {path}")

class CkanMandatoryArgumentError(Exception):
    def __init__(self, action_name: str, attribute_name: str):
        super().__init__(f"Argument '{attribute_name}' is required for {action_name}")

class MandatoryAttributeError(Exception):
    def __init__(self, object_type: str, attribute_name: str):
        super().__init__(f"Attribute '{attribute_name}' is required for {object_type} to initiate builder")

class MissingIdError(Exception):
    def __init__(self, object_type: str, object_name):
        super().__init__(f"Attribute 'id' is required for {object_type} '{object_name}' to update CKAN map")

class CkanServerError(Exception):
    def __init__(self, ckan, response: requests.Response, msg:str, display_request:bool=True):
        super().__init__(msg)
        self.response = response
        self.status_code = response.status_code
        if display_request:
            ckan._error_print_debug_response(response)

    def __str__(self):
        return f"Server code [{self.status_code}]: " + super().__str__()

class DataStoreNotFoundError(Exception):
    def __init__(self, resource_id:str, error_message: str):
        super().__init__(f"DataStore not found for resource id {resource_id}. This could mean the DataStore was not initialized. Server message: {error_message}")

class DuplicateNameError(Exception):
    def __init__(self, object_type:str, names:Iterable[str]):
        super().__init__(f"Duplicate names were found for {object_type}: {','.join(names)}")

class ForbiddenNameError(Exception):
    def __init__(self, object_type:str, names:Iterable[str]):
        super().__init__(f"Forbidden name for {object_type}: {','.join(names)}")

class IntegrityError(Exception):
    pass

class ReadOnlyError(Exception):
    def __init__(self):
        super().__init__("Mode is set to read only. Please set the read_only flag to False.")

class AdminFeatureLockedError(Exception):
    def __init__(self):
        super().__init__("Admin features are locked. Please set the enable_admin flag to True.")

class NotMappedObjectNameError(Exception):
    pass

class UnexpectedError(RuntimeError):
    pass

class UrlError(Exception):
    pass

class MaxRequestsCountError(Exception):
    def __init__(self):
        super().__init__("Maximum requests count was reached.")

class CkanArgumentError(Exception):
    def __init__(self, api_name:str, argument_name:str):
        super().__init__(f"Argument {argument_name} is not supported by API {api_name}.")

class ArgumentError(Exception):
    pass

class SearchAllNoCountsError(ArgumentError):
    def __init__(self, api_name:str, argument_name_value:str=None):
        if argument_name_value is None:
            super().__init__(f"{api_name} must parse results to compute the number of rows returned. Argument return_df=False is incompatible with multi-request mode search_all=True")
        else:
            super().__init__(f"{api_name} must parse results to compute the number of rows returned. Arguments return_df=False and {argument_name_value} are incompatible with multi-request mode search_all=True")

class FunctionMissingArgumentError(Exception):
    def __init__(self, function_name:str, argument_name:str):
        super().__init__(f"Argument {argument_name} is mandatory for function {function_name}.")

class NoDefaultView(Exception):
    def __init__(self, resource_format:str):
        super().__init__(f"No default view defined for resource format {resource_format}")

class ExternalUrlLockedError(Exception):
    def __init__(self, url:str):
        super().__init__(f"Downloading external urls is blocked by parameter download_external_urls (url {url}). Run unlock_external_url_resource_download to enable this feature.")

class NoCAVerificationError(Exception):
    def __init__(self):
        super().__init__("The CA verification cannot be disabled. To unlock this feature, run unlock_no_ca to enable this feature. Warning: Only allow in a local environment!")

class RequestError(Exception):
    pass

class RequirementError(Exception):
    pass

class FileFormatRequirementError(RequirementError):
    def __init__(self, requirement:str, file_format:str):
        super().__init__(f"The package {requirement} is required to support this file format ({file_format}).")

# PostGIS
class UnknownTargetCRSError(RequirementError):
    def __init__(self, source_crs, context:str):
        super().__init__(f"Unknown destination CRS (source={source_crs}) for {context}.")




#!python3
# -*- coding: utf-8 -*-
"""
Tests to perform after the example package was uploaded
"""
from typing import Tuple
import os
import re
import getpass
import json

import pandas as pd
import numpy as np

from ckanapi_harvesters.auxiliary import CkanMap
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_action import CkanNotFoundError, CkanSqlCapabilityError

from ckanapi_harvesters.builder.example import example_package_xls
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))


def run(ckan:CkanApi = None):
    BuilderPackage.unlock_external_code_execution()  # comment to test if the safety feature is enabled

    mdl = BuilderPackage.from_excel(example_package_xls)
    ckan = mdl.init_ckan(ckan)
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    # ckan.apikey.clear()  # uncomment for specific test

    capability = ckan.test_sql_capabilities()
    try:
        ckan.api_help_show("datastore_search_sql")
    except CkanNotFoundError:
        print("No datastore_search_sql help")
    users_table_id = mdl.get_or_query_resource_id(ckan, "users.csv")
    try:
        ckan.datastore_search_sql(f'SELECT * FROM "{users_table_id}"')
    except CkanSqlCapabilityError:
        print("datastore_search_sql is not accessible")

    print("Tests done.")


if __name__ == '__main__':
    ckan = CkanApi(None)
    ckan.initialize_from_cli_args()
    run(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Abstract class to implement specific builders from code
"""
from abc import ABC
from typing import List

from ckanapi_harvesters.auxiliary.ckan_model import CkanState
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.builder.builder_package import BuilderPackage

class SpecificBuilderABC(BuilderPackage, ABC):
    def __init__(self, ckan:CkanApi, package_name:str, organization_name:str, *,
                 title: str = None, description: str = None, private: bool = None, state: CkanState = None,
                 version: str = None,
                 url: str = None, tags: List[str] = None,
                 license_name:str=None):
        super().__init__(package_name=package_name, title=title,
                         description=description, private=private, state=state, version=version, url=url,
                         tags=tags, organization_name=organization_name, license_name=license_name)
        self.ckan_builder.from_ckan(ckan)




#!python3
# -*- coding: utf-8 -*-
"""
Code to initiate a DataStore defined by a large number of files to concatenate into one table.
This concrete implementation is linked to the file system.
"""
from typing import Dict, List, Collection, Any, Tuple, Generator, Union, Set
from collections import OrderedDict
from warnings import warn
import glob
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise
from ckanapi_harvesters.builder.mapper_datastore import DataSchemeConversion
# from ckanapi_harvesters.auxiliary.path import list_files_scandir
from ckanapi_harvesters.builder.builder_errors import ResourceFileNotExistMessage
from ckanapi_harvesters.builder.builder_resource_datastore_multi_abc import BuilderDataStoreMultiABC
from ckanapi_harvesters.builder.builder_resource_datastore_multi_abc import datastore_multi_apply_last_condition_intermediary
from ckanapi_harvesters.builder.builder_field import BuilderField
from ckanapi_harvesters.auxiliary.ckan_model import CkanField, CkanResourceInfo, UpsertChoice
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element
from ckanapi_harvesters.builder.mapper_datastore_multi import RequestMapperABC, RequestFileMapperABC
from ckanapi_harvesters.builder.mapper_datastore_multi import default_file_mapper_from_primary_key
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreFile
from ckanapi_harvesters.harvesters.harvester_abc import TableHarvesterABC
from ckanapi_harvesters.harvesters.harvester_init import init_table_harvester_from_options_string
from ckanapi_harvesters.builder.builder_resource_datastore_multi_folder import BuilderDataStoreFolder


class BuilderDataStoreHarvester(BuilderDataStoreFolder):
    def __init__(self, *, file_query_list: List[Tuple[str,dict]]=None, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, dir_name:str=None, file_url_attr:str=None, options_string:str=None, base_dir:str=None):
        super().__init__(file_query_list=file_query_list, dir_name=dir_name,
                         name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        self.options_string = options_string
        self.enable_multi_threaded_upload = False
        # specific attributes
        self.file_url_attr:Union[str,None] = file_url_attr
        self._harvester: Union[TableHarvesterABC,None] = None
        if self.options_string is not None and len(self.options_string) > 0:
            self._apply_options(base_dir=base_dir)

    @property
    def harvester(self) -> Union[TableHarvesterABC,None]:
        return self._harvester
    @harvester.setter
    def harvester(self, harvester: Union[TableHarvesterABC,None]):
        assert_or_raise(self._harvester is None, RuntimeError("You can only set the harvester once"))
        self._harvester = harvester
        self._apply_harvester_metadata()

    def _apply_options(self, base_dir: str = None):
        self.harvester = init_table_harvester_from_options_string(self.options_string, file_url_attr=self.file_url_attr, base_dir=base_dir)

    def init_options_from_ckan(self, ckan:CkanApi) -> None:
        super().init_options_from_ckan(ckan)
        self.harvester.update_from_ckan(ckan)

    def _apply_harvester_metadata(self, base_dir:str=None):
        self.dir_name = self.name  # by default, take the resource name
        if self.harvester.params.output_dir is not None:
            self.dir_name = self.harvester.params.output_dir
        if self.harvester.params.enable_download is not None:
            self.enable_download = self.harvester.params.enable_download
        # import default metadata
        table_metadata = self.harvester.clean_table_metadata()
        if self.df_mapper.df_upload_fun is None:
            self.df_mapper.df_upload_fun = self.harvester.get_default_df_upload_fun()
        if self.data_cleaner_upload is None:
            self.data_cleaner_upload = self.harvester.get_default_data_cleaner()
        if self.primary_key is None:
            self.primary_key = self.harvester.get_default_primary_key()
        if self.indexes is None:
            self.indexes = table_metadata.indexes
        if self.description is None:
            self.description = table_metadata.description
        if self.format is None:
            self.format = "CSV"
        if table_metadata.fields is not None:
            if self.field_builders is None:
                self.field_builders = OrderedDict()
            for field_name, field_metadata in table_metadata.fields.items():
                if field_name in self.field_builders.keys():
                    field_builder = self.field_builders[field_name]
                    if field_builder.type_override is None:
                        field_builder.type_override = field_metadata.data_type
                else:
                    field_builder = BuilderField(name=field_metadata.name,
                                                 type_override=field_metadata.data_type)
                if field_builder.label is None:
                    field_builder.label = field_metadata.label
                if field_builder.description is None:
                    field_builder.description = field_metadata.description
                if field_builder.uniquekey is None:
                    field_builder.uniquekey = field_metadata.uniquekey or (table_metadata.unique_keys is not None and field_name in table_metadata.unique_keys)
                if field_builder.is_index is None:
                    field_builder.is_index = field_metadata.is_index
                if field_builder.notnull is None:
                    field_builder.notnull = field_metadata.notnull
                field_builder.internal_attrs = field_metadata.internal_attrs.merge(field_builder.internal_attrs)
                self.field_builders[field_name] = field_builder
        if table_metadata.unique_keys is not None and len(table_metadata.unique_keys) > 0:
            if self.field_builders is None:
                self.field_builders = OrderedDict()
            for field_name in table_metadata.unique_keys:
                if field_name in self.field_builders.keys():
                    field_builder = self.field_builders[field_name]
                    if field_builder.uniquekey is None:
                        field_builder.uniquekey = True
                else:
                    pass  # because we do not know the data type
                    # field_builder = BuilderField(name=field_name)
                    # field_builder.uniquekey = field_name
                    # self.field_builders[field_name] = field_builder

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreHarvester()
        super().copy(dest=dest)
        dest.file_url_attr = self.file_url_attr
        dest.harvester = self.harvester
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None) -> None:
        super()._load_from_df_row(row=row)
        self.df_mapper = default_file_mapper_from_primary_key(self.primary_key)
        self.dir_name = ""
        self.file_url_attr: str = _string_from_element(row["file/url"])
        if self.options_string is not None and len(self.options_string) > 0:
            self._apply_options(base_dir=base_dir)

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.file_url_attr
        return d

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from Harvester"

    @staticmethod
    def from_file_datastore(resource_file: BuilderDataStoreFile,
                            *, dir_name:str=None, primary_key:List[str]=None,
                            file_query_list:Collection[Tuple[str,dict]]=None) -> "BuilderDataStoreHarvester":
        """
        Do not initialize a BuilderDataStoreHarvester with this method. Rather initialize a new instance of the class.

        :raises NotImplementedError:
        """
        raise NotImplementedError("This method must not be called for a DataStore from Harvester. Rather initialize a new BuilderDataStoreHarvester.")


    ## upload is specific to this class ---------------------------------------------------
    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        return self.harvester.check_connection()

    def get_sample_file_path(self, resources_base_dir:str, file_index:int=0) -> Union[Any,None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        return self.local_file_list[file_index]

    def load_local_df(self, file: str, *, upload_alter:bool=True, fields:OrderedDict[str,CkanField]=None) -> pd.DataFrame:
        # self.sample_data_source = resolve_rel_path(resources_base_dir, self.dir_name, file, field=f"File/URL of resource {self.name}")
        self.sample_data_source = file
        data_local = self.harvester.query_data(query=file)
        if upload_alter:
            df_upload = self.df_mapper.df_upload_alter(data_local, self.sample_data_source, fields=self._get_fields_info())
            return df_upload
        else:
            raise RuntimeError("upload_alter must be True for a DataStore from Harvester.")

    def get_local_file_generator(self, resources_base_dir:str, **kwargs) -> Generator[Any, None, None]:
        self.list_local_files(resources_base_dir=resources_base_dir)
        for query in self.local_file_list:
            yield query

    def list_local_files(self, resources_base_dir:str, cancel_if_present:bool=True) -> List[Any]:
        if cancel_if_present and self.local_file_list is not None:
            return self.local_file_list
        self.local_file_list = self.harvester.list_queries(new_connection=not cancel_if_present)
        return self.local_file_list

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        return self.list_local_files(resources_base_dir=resources_base_dir, cancel_if_present=cancel_if_present)

    def get_local_df_generator(self, resources_base_dir:str, *, fields:OrderedDict[str,CkanField], **kwargs) -> Generator[pd.DataFrame, None, None]:
        return super().get_local_df_generator(resources_base_dir=resources_base_dir, fields=fields, **kwargs)

    def get_local_file_len(self) -> int:
        if self.local_file_list is None:
            raise RuntimeError("You must call list_local_files first")
        return len(self.local_file_list)

    # def patch_request(self, ckan: CkanApi, package_id: str, *,
    #                   df_upload: pd.DataFrame=None, reupload: bool = None, resources_base_dir:str=None) -> CkanResourceInfo:
    #     # apply same treatments as super method to determine df_upload
    #     if reupload is None: reupload = self.reupload_on_update
    #     if df_upload is None:
    #         if not reupload:
    #             resource_id = ckan.map.get_resource_id(self.name, self.package_name, error_not_mapped=False)
    #             if resource_id is not None:
    #                 fields = ckan.get_datastore_fields_or_request(resource_id, error_not_found=False)
    #             else:
    #                 fields = None
    #         else:
    #             fields = None
    #         df_upload = self.load_sample_df(resources_base_dir=resources_base_dir, upload_alter=True, fields=fields)
    #     return super().patch_request(ckan, package_id, df_upload=df_upload, reupload=reupload, resources_base_dir=resources_base_dir)

    # def upload_request_full(self, ckan:CkanApi, resources_base_dir:str, *,
    #                         method:UpsertChoice=UpsertChoice.Upsert,
    #                         threads:int=1, external_stop_event=None,
    #                         only_missing:bool=False,
    #                         start_index:int=0, end_index:int=None) -> None:
    #     resource_id = ckan.map.get_resource_id(self.name, self.package_name, error_not_mapped=False)
    #     if resource_id is not None:
    #         fields = ckan.get_datastore_fields_or_request(resource_id, error_not_found=False)
    #     else:
    #         fields = None
    #     super().upload_request_full(ckan=ckan, resources_base_dir=resources_base_dir,
    #                                 threads=threads, external_stop_event=external_stop_event,
    #                                 start_index=start_index, end_index=end_index,
    #                                 method=method, fields=fields)

    def upsert_request_df(self, ckan: CkanApi, df_upload:pd.DataFrame,
                          method:UpsertChoice=UpsertChoice.Upsert,
                          apply_last_condition:bool=None, always_last_condition:bool=None) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Call to ckan datastore_upsert.
        Before sending the DataFrame, a call to df_upload_alter is made.
        This implementation optionally checks for the last line of the DataFrame based on the first columns of the primary key.

        :param ckan:
        :param df_upload:
        :param method:
        :return:
        """

        # resource_id = self.get_or_query_resource_id(ckan, error_not_found=True)
        # df_upload_transformed = self.df_mapper.df_upload_alter(df_upload)
        # ret_df = ckan.datastore_upsert(df_upload_transformed, resource_id, method=method,
        #                                apply_last_condition=apply_last_condition,
        #                                always_last_condition=always_last_condition)
        # return df_upload_transformed, ret_df

        if apply_last_condition is None:
            apply_last_condition = True  # datastore_multi_apply_last_condition_intermediary
        resource_id = self.get_or_query_resource_id(ckan=ckan, error_not_found=True)
        df_upload_local = df_upload
        df_upload_transformed = self.df_mapper.df_upload_alter(df_upload_local, fields=self._get_fields_info())
        file_query = self.df_mapper.get_file_query_of_df(df_upload_transformed)
        if file_query is not None:
            i_restart, upload_needed, row_count, df_row = self.df_mapper.last_inserted_index_request(ckan=ckan,
                                     resource_id=resource_id, df_upload=df_upload_transformed, file_query=file_query)
        else:
            i_restart, upload_needed, row_count, df_row = 0, True, -1, None
        if upload_needed:
            if i_restart > 0 and ckan.params.verbose_extra:
                print(f"Starting transfer from index {i_restart}")
            ret_df = ckan.datastore_upsert(df_upload_transformed.iloc[i_restart:], resource_id, method=method,
                                           apply_last_condition=apply_last_condition,
                                           always_last_condition=always_last_condition, data_cleaner=self.data_cleaner_upload)
        elif 0 <= row_count and row_count < len(df_row):
            msg = f"Sending full dataframe because is was shorter on server side"
            warn(msg)
            ret_df = ckan.datastore_upsert(df_upload_transformed, resource_id, method=method,
                                           apply_last_condition=apply_last_condition,
                                           always_last_condition=always_last_condition, data_cleaner=self.data_cleaner_upload)
        else:
            if ckan.params.verbose_extra:
                print(f"File up to date on server side")
            ret_df = None
        return df_upload_transformed, ret_df





#!python3
# -*- coding: utf-8 -*-
"""
Section of the package dedicated to the initialization of a CKAN package
"""

import os

# usage shortcuts
self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
example_package_dir = os.path.join(self_dir, "package")
from ..builder_package import example_package_xls

from . import builder_example
from . import builder_example_aux_fun
from . import builder_example_generate_data
from . import builder_example_patch_upload
from . import builder_example_tests
from . import builder_example_policy
from . import builder_example_download




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Set, Union, Tuple, Dict
from warnings import warn
import json
import os
import copy

import requests
from requests.auth import AuthBase

from ckanapi_harvesters.auxiliary.ckan_configuration import allow_policy_from_url, download_external_resource_urls
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.auxiliary.urls import is_valid_url
from ckanapi_harvesters.auxiliary.path import path_rel_to_dir
from ckanapi_harvesters.auxiliary.ckan_errors import ExternalUrlLockedError
from ckanapi_harvesters.policies import POLICY_FILE_FORMAT_VERSION
from ckanapi_harvesters.policies.data_format_policy_errors import (DataPolicyError, UnsupportedPolicyVersionError,
                                                                   _policy_msg, ErrorCount, ErrorLevel, UrlPolicyLockedError)
from ckanapi_harvesters.policies.data_format_policy_defs import StringMatchMode, newline_char
from ckanapi_harvesters.policies.data_format_policy_defs import ListChoiceMode, StringValueSpecification
from ckanapi_harvesters.policies.data_format_policy_abc import DataPolicyABC
from ckanapi_harvesters.policies.data_format_policy_lists import ValueListPolicy, GroupedValueListPolicy, SingleValueListPolicy
from ckanapi_harvesters.policies.data_format_policy_tag_groups import TagListPolicy, TagGroupsListPolicy
from ckanapi_harvesters.policies.data_format_policy_custom_fields import CustomFieldSpecification, CustomFieldsPolicy
from ckanapi_harvesters.auxiliary.ckan_model import CkanPackageInfo, CkanConfigurableObjectABC



class CkanPackageDataFormatPolicy(DataPolicyABC):
    """
    Main class to define data format policy for package metadata
    """
    default_to_json_reduced_size:bool = False

    def __init__(self, label:str=None, description:str=None,
                 package_tags:TagGroupsListPolicy=None, package_custom_fields:CustomFieldsPolicy=None,
                 package_mandatory_attributes:Set[str]=None, resource_mandatory_attributes:Set[str]=None,
                 datastore_fields_mandatory_attributes:Set[str]=None, resource_format:SingleValueListPolicy=None):
        super().__init__()
        if label is None:
            label = "Policy"
        if description is None:
            description = ""
        if isinstance(package_mandatory_attributes, str):
            package_mandatory_attributes = set(ckan_tags_sep.split(package_mandatory_attributes))
        if isinstance(resource_mandatory_attributes, str):
            resource_mandatory_attributes = set(ckan_tags_sep.split(resource_mandatory_attributes))
        if isinstance(datastore_fields_mandatory_attributes, str):
            datastore_fields_mandatory_attributes = set(ckan_tags_sep.split(datastore_fields_mandatory_attributes))
        self.label: str = label
        self.description: str = description
        self.package_tags:TagGroupsListPolicy = package_tags
        self.package_custom_fields: CustomFieldsPolicy = package_custom_fields
        self.package_mandatory_attributes:Set[str] = package_mandatory_attributes
        self.resource_mandatory_attributes:Set[str] = resource_mandatory_attributes
        self.datastore_fields_mandatory_attributes:Set[str] = datastore_fields_mandatory_attributes
        self.resource_format:SingleValueListPolicy = resource_format
        self.file_format_version:Union[str,None] = None
        self.source_file: Union[str,None] = None

    def __copy__(self):
        return self.copy()

    def copy(self) -> "CkanPackageDataFormatPolicy":
        dest = CkanPackageDataFormatPolicy()
        dest.label = self.label
        dest.description = self.description
        dest.package_tags = copy.deepcopy(self.package_tags)
        dest.package_custom_fields = copy.deepcopy(self.package_custom_fields)
        dest.package_mandatory_attributes = copy.deepcopy(self.package_mandatory_attributes)
        dest.resource_mandatory_attributes = copy.deepcopy(self.resource_mandatory_attributes)
        dest.datastore_fields_mandatory_attributes = copy.deepcopy(self.datastore_fields_mandatory_attributes)
        dest.resource_format = copy.deepcopy(self.resource_format)
        dest.file_format_version = self.file_format_version
        dest.source_file = self.source_file
        dest.error_level = self.error_level
        return dest

    def to_dict(self, *, sets_as_lists:bool=True) -> dict:
        d = {"info": {"file_format_version": POLICY_FILE_FORMAT_VERSION,
                      "label": self.label,
                      "description": self.description,
                      },}
        if self.package_tags is not None:
            d["package_tags_policy"] = self.package_tags.to_dict()
        if self.package_custom_fields is not None:
            d["package_custom_fields_policy"] = self.package_custom_fields.to_dict()
        if self.package_mandatory_attributes is not None:
            set_object = self.package_mandatory_attributes
            if sets_as_lists:
                set_object = sorted(list(set_object))
            d["package_mandatory_attributes"] = set_object
        if self.resource_mandatory_attributes is not None:
            set_object = self.resource_mandatory_attributes
            if sets_as_lists:
                set_object = sorted(list(set_object))
            d["resource_mandatory_attributes"] = set_object
        if self.datastore_fields_mandatory_attributes is not None:
            set_object = self.datastore_fields_mandatory_attributes
            if sets_as_lists:
                set_object = sorted(list(set_object))
            d["datastore_fields_mandatory_attributes"] = set_object
        if self.resource_format is not None:
            d["resource_format_policy"] = self.resource_format.to_dict()
        d.update(super().to_dict())
        return {"ckan_package_policy": d}

    @staticmethod
    def from_dict(d:dict) -> "CkanPackageDataFormatPolicy":
        obj = CkanPackageDataFormatPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        d = d["ckan_package_policy"]
        super()._load_from_dict(d)
        self.file_format_version = d["info"]["file_format_version"]
        if not self.file_format_version == POLICY_FILE_FORMAT_VERSION:
            raise UnsupportedPolicyVersionError(self.file_format_version)
        self.label = d["info"]["label"]
        self.description = d["info"]["description"]
        # for package tags management, see also tags and vocabularies in the CKAN API documentation. Here, tags groups are the equivalent of tag vocabularies.
        self.package_tags = TagGroupsListPolicy.from_dict(d["package_tags_policy"]) if "package_tags_policy" in d.keys() else None
        self.package_custom_fields = CustomFieldsPolicy.from_dict(d["package_custom_fields_policy"]) if "package_custom_fields_policy" in d.keys() else None
        self.package_mandatory_attributes = set(d["package_mandatory_attributes"]) if "package_mandatory_attributes" in d.keys() else None
        self.resource_mandatory_attributes = set(d["resource_mandatory_attributes"]) if "resource_mandatory_attributes" in d.keys() else None
        self.datastore_fields_mandatory_attributes = set(d["datastore_fields_mandatory_attributes"]) if "datastore_fields_mandatory_attributes" in d.keys() else None
        self.resource_format = SingleValueListPolicy.from_dict(d["resource_format_policy"]) if "resource_format_policy" in d.keys() else None

    def to_json(self, json_file:str, reduced_size:bool=None) -> None:
        if reduced_size is None:
            reduced_size = self.default_to_json_reduced_size
        policy_dict = self.to_dict()
        with open(json_file, "w", encoding="utf-8") as json_file:
            if reduced_size:
                json.dump(policy_dict, json_file, ensure_ascii=False)
            else:
                json.dump(policy_dict, json_file, ensure_ascii=False, indent=4)

    def to_jsons(self, reduced_size:bool=None) -> str:
        if reduced_size is None:
            reduced_size = self.default_to_json_reduced_size
        policy_dict = self.to_dict()
        if reduced_size:
            # do not include spaces and line endings (not human-readable format)
            return json.dumps(policy_dict, ensure_ascii=False)
        else:
            return json.dumps(policy_dict, indent=4, ensure_ascii=False)

    @staticmethod
    def from_jsons(stream:str, *,
                   source_file:str=None, load_error:bool=True) -> Union["CkanPackageDataFormatPolicy", None]:
        try:
            policy_dict = json.loads(stream)
            obj = CkanPackageDataFormatPolicy.from_dict(policy_dict)
        except Exception as e:
            if load_error:
                raise e from e
            else:
                msg = f"Could not load policy (JSON error): {str(e)}"
                warn(msg)
                return None
        obj.source_file = source_file
        return obj

    @staticmethod
    def from_json(policy_file:str, *, base_dir:str=None,
                  headers:dict=None, proxies:dict=None, auth:Union[AuthBase, Tuple[str,str]]=None, verify:Union[bool,str,None]=None,
                  error_not_found:bool=True) -> Union["CkanPackageDataFormatPolicy",None]:
        policy_dict = None
        if is_valid_url(policy_file):
            if not allow_policy_from_url:
                raise UrlPolicyLockedError(policy_file)
            # if (not download_external_resource_urls) and (not ckan.is_url_internal(policy_file)):  # ckan: unknown
            #     raise ExternalUrlLockedError(policy_file)
            response = requests.get(policy_file, headers=headers, proxies=proxies, auth=auth, verify=verify)
            if response.status_code != 200 and error_not_found:
                raise FileNotFoundError(policy_file)
            policy_dict = json.loads(response.content.decode())
        else:
            policy_file = path_rel_to_dir(policy_file, base_dir)
            if not os.path.isfile(policy_file) and not error_not_found:
                return None
            with open(policy_file, "r") as f:
                policy_dict = json.load(f)
        obj = CkanPackageDataFormatPolicy.from_dict(policy_dict)
        obj.source_file = policy_file
        return obj

    def _enforce_attributes_list(self, value:CkanConfigurableObjectABC, spec:Set[str], *, context:str, verbose: bool, buffer:List[DataPolicyError]):
        extra_spec = spec - value.configurable_attributes
        if len(extra_spec) > 0:
            raise KeyError("These attributes do not exist for " + value.get_resource_type() + ": " + ",".join(extra_spec) + ". Allowed attributes: " + str(value.configurable_attributes))
        current_attributes = {name for name in value.configurable_attributes if getattr(value, name) is not None}
        missing_attributes = set(spec) - current_attributes
        if missing_attributes:
            msg = DataPolicyError(context, self.error_level, f"Mandatory attributes were not found: {', '.join(missing_attributes)}")
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            return False
        else:
            return True

    def enforce(self, values: CkanPackageInfo, *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        package_info = values
        success = True
        if context is None:
            context = "Package " + package_info.name
        if self.package_tags is not None:
            success &= self.package_tags.enforce(package_info.tags, context=context + " / package tags", verbose=verbose, buffer=buffer)
        if self.package_custom_fields is not None:
            success &= self.package_custom_fields.enforce(package_info.custom_fields, context=context, verbose=verbose, buffer=buffer)
        if self.package_mandatory_attributes is not None:
            success &= self._enforce_attributes_list(package_info, self.package_mandatory_attributes, context=context, verbose=verbose, buffer=buffer)
        for resource_info in package_info.package_resources.values():
            resource_context = context + " / resource " + resource_info.name
            if self.resource_format is not None:
                resource_format_context = resource_context + " / resource format"
                success &= self.resource_format.enforce(resource_info.format, context=resource_format_context, verbose=verbose, buffer=buffer)
            if self.resource_mandatory_attributes is not None:
                success &= self._enforce_attributes_list(resource_info, self.resource_mandatory_attributes, context=resource_context, verbose=verbose, buffer=buffer)
            if self.datastore_fields_mandatory_attributes is not None and resource_info.datastore_info is not None:
                for field_info in resource_info.datastore_info.fields:
                    field_context = resource_context + " / field " + field_info.name
                    success &= self._enforce_attributes_list(field_info, self.datastore_fields_mandatory_attributes, context=field_context, verbose=verbose, buffer=buffer)
        return success

    def policy_check_package(self, package_info: CkanPackageInfo, *, package_buffer:List[DataPolicyError]=None,
                             display_message:bool=True, raise_error:bool=False) -> bool:
        """
        Main entry-point to check the policy rules against the package.

        :param package_info: package and resources metadata
        :param package_buffer: you can specify a list object to indirectly obtain the detailed list of error messages.
        The keys of this dictionary are the package names.
        :param display_message: option to display the messages in the command line
        :param raise_error: option to raise an exception if any rule with a high error level is encountered
        :return: True if no error was encountered
        """
        if package_buffer is None:
            package_buffer: List[DataPolicyError] = []
        context = "Package " + package_info.name
        success = self.enforce(package_info, context=context, verbose=True, buffer=package_buffer)
        error_count = ErrorCount(package_buffer)
        # consistency check
        if success:
            assert(error_count.total == 0)
        else:
            assert(error_count.total > 0)
        # command-line output
        if display_message:
            if success:
                print("Package '" + package_info.name + "' passed all tests")
            else:
                print("Package '" + package_info.name + "': " + error_count.error_count_message() + ":")
                print('\n'.join([error_message.message for error_message in package_buffer]))
        # raise error after all this
        if raise_error and error_count.error > 0:
            raise DataPolicyError(context, ErrorLevel.Error, error_count.error_count_message())
        return success









#!python3
# -*- coding: utf-8 -*-
"""
Classes to read specific file formats to load DataStore DataFrame/records from a system file
"""

from . import file_format_abc
from . import csv_format
from . import shp_format
from . import file_format_init



#!python3
# -*- coding: utf-8 -*-
"""
Harvester parameters. The base names of the parameters are shared between harvesters.
"""
from typing import Union, Tuple, List, Any, Callable
from collections import OrderedDict
from abc import ABC, abstractmethod
import argparse
import shlex
from warnings import warn
import copy

import pandas as pd
from requests.auth import AuthBase

from ckanapi_harvesters.auxiliary.ckan_configuration import default_ckan_has_postgis, default_ckan_target_epsg
from ckanapi_harvesters.auxiliary.ckan_configuration import unlock_external_url_resource_download, allow_no_ca, unlock_no_ca
from ckanapi_harvesters.auxiliary.ckan_errors import NoCAVerificationError
from ckanapi_harvesters.auxiliary.ckan_auxiliary import ca_file_rel_to_dir, assert_or_raise
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_api_key import ApiKey
from ckanapi_harvesters.auxiliary.login import Login, SSHLogin
from ckanapi_harvesters.harvesters.harvester_errors import HarvestMethodRequiredError

harvester_enforce_ca_verification: bool = False


class DatabaseParams:
    """
    Class representing parameters to connect to a database.
    This class mangages the connection parameters such as proxy and CA.
    It also manages authentication parameters.
    """
    def __init__(self, source: "DatabaseParams" = None):
        self.options_string: Union[str,None] = None
        self.file_url_attr: Union[str, None] = None
        self.base_dir: Union[str,None] = None
        self.harvest_method: str = ""
        self._proxy_config: ProxyConfig = ProxyConfig()
        self._verify_ca: Union[str, bool, None] = None
        self._verify_ca_src: Union[str, None] = None
        self.timeout: Union[float, None] = None
        self.host: Union[str, None] = None
        self.port: Union[int, None] = None
        self.auth_url_suffix: Union[str, None] = None
        self.auth_url: Union[str, None] = None
        self.url: Union[str, None] = None
        self.apikey: ApiKey = ApiKey()
        self.login: Login = Login()
        self.database: Union[str, None] = None
        self.verbose_harvester: bool = True
        self.ckan_postgis: Union[bool,None] = default_ckan_has_postgis
        self.ckan_default_target_epsg:Union[int,None] = default_ckan_target_epsg
        if source is not None:
            source.copy(dest=self)

    @abstractmethod
    def copy(self, *, dest=None):
        dest.options_string = self.options_string
        dest.file_url_attr = self.file_url_attr
        dest.base_dir = self.base_dir
        dest.harvest_method = self.harvest_method
        dest._proxy_config = self._proxy_config
        dest._verify_ca = self._verify_ca
        dest._verify_ca_src = self._verify_ca_src
        dest.timeout = self.timeout
        dest.host = self.host
        dest.port = self.port
        dest.auth_url_suffix = self.auth_url_suffix
        dest.auth_url = self.auth_url
        dest.url = self.url
        dest.apikey = self.apikey
        dest.login = self.login
        dest.database = self.database
        dest.verbose_harvester = self.verbose_harvester
        dest.ckan_postgis = self.ckan_postgis
        return dest

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="Harvester parameters")
        parser.add_argument("--harvester", type=str,
                            help="Type of harvester to use", required=True)
        ProxyConfig._setup_cli_proxy_parser(parser)  # add arguments --proxy --http-proxy --https-proxy --no-proxy --proxy-auth-file
        parser.add_argument("--ca", type=str,
                            help="Server CA certificate location (.pem file)")
        parser.add_argument("--timeout", type=float,
                            help="Server timeout (seconds)")
        parser.add_argument("--host", type=str,
                            help="Host for queries")
        parser.add_argument("--port", type=int,
                            help="Port for queries")
        parser.add_argument("--auth-url-suffix", type=str,
                            help="URL suffix used to authenticate user")
        parser.add_argument("--auth-url", type=str,
                            help="URL to authenticate user")
        parser.add_argument("--url", type=str,
                            help="Base URL for queries")
        ApiKey._setup_cli_parser(parser)  # add arguments --apikey-file --apikey
        Login._setup_cli_parser(parser)  # add argument --login-file
        parser.add_argument("-v", "--verbose",
                            help="Option to set verbosity", action="store_true", default=False)
        parser.add_argument("--database", type=str,
                            help="Database name")
        parser.add_argument("--ckan-postgis", action="store_true",
                            help="Option to use CKAN with PostGIS geometric types")  # default=default_ckan_has_postgis
        parser.add_argument("--ckan-epsg", type=int,
                            help="Default EPSG for CKAN", default=default_ckan_target_epsg)
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        self.harvest_method = args.harvester
        proxy_config = ProxyConfig.from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                                 default_proxies=default_proxies, proxy_headers=proxy_headers)
        if proxy_config is not None:
            self._proxy_config = proxy_config
        ca_cert = args.ca
        verify_ca, self._verify_ca_src = ca_file_rel_to_dir(ca_cert, base_dir=base_dir)
        self.set_verify_ca(verify_ca)
        self.timeout = args.timeout
        self.host = args.host
        self.port = args.port
        self.auth_url_suffix = args.auth_url_suffix
        self.auth_url = args.auth_url
        self.url = args.url
        self.apikey._cli_args_apply(args, base_dir=base_dir, error_not_found=error_not_found)
        self.login._cli_args_apply(args, base_dir=base_dir, error_not_found=error_not_found)
        self.database = args.database
        if args.verbose is not None:
            self.verbose_harvester = args.verbose
        if args.ckan_postgis:
            self.ckan_postgis = args.ckan_postgis
        if args.ckan_epsg:
            self.ckan_default_target_epsg = args.ckan_epsg

    def _update_from_ckan(self, ckan):
        # aim: make these values accessible to the harvester algorithms (for the rest, Harvesters are independent of CkanApi)
        if self.ckan_postgis is None:
            self.ckan_postgis = ckan.params.ckan_has_postgis
        if self.ckan_default_target_epsg is None:
            self.ckan_default_target_epsg = ckan.params.ckan_default_target_epsg

    @staticmethod
    def parse_harvest_method(options_string: str) -> str:
        # parser = DatabaseParams.setup_cli_harvester_parser()
        parser = argparse.ArgumentParser(description="Harvester selection")
        parser.add_argument("--harvester", type=str,
                            help="Type of harvester to use", required=True)
        args, _ = parser.parse_known_args(shlex.split(options_string))
        assert_or_raise(args.harvester is not None, HarvestMethodRequiredError())
        return args.harvester.lower().strip()

    def parse_options_string(self, options_string: str, *, base_dir: str = None, file_url_attr: str=None,
                             parser:argparse.ArgumentParser=None):
        self.file_url_attr = file_url_attr
        parser = self.setup_cli_harvester_parser(parser)
        args, _ = parser.parse_known_args(shlex.split(options_string))
        self.options_string = options_string
        self.base_dir = base_dir
        self.initialize_from_cli_args(args, base_dir=base_dir)

    @property
    def proxies(self) -> dict:
        return self._proxy_config.proxies

    @proxies.setter
    def proxies(self, proxies: dict) -> None:
        self._proxy_config.proxies = proxies

    @property
    def proxy_string(self) -> str:
        return self._proxy_config.proxy_string

    @proxy_string.setter
    def proxy_string(self, proxies: str) -> None:
        self._proxy_config.proxy_string = proxies

    @property
    def proxy_auth(self) -> Union[AuthBase, Tuple[str, str]]:
        return self._proxy_config.proxy_auth

    @proxy_auth.setter
    def proxy_auth(self, proxy_auth: Union[AuthBase, Tuple[str, str]]) -> None:
        self._proxy_config.proxy_auth = proxy_auth

    @property
    def verify_ca(self) -> Union[bool, str, None]:
        return self._verify_ca

    def set_verify_ca(self, ca_cert: Union[bool, str, None], enforce_ca_safety: bool = None) -> None:
        if enforce_ca_safety is None:
            enforce_ca_safety = harvester_enforce_ca_verification
        if ca_cert is not None and isinstance(ca_cert, bool) and not ca_cert:
            if enforce_ca_safety and not allow_no_ca:
                raise NoCAVerificationError()
            else:
                msg = "CA verification has been disabled. Only allow in a local environment!"
                warn(msg)
        self._verify_ca = ca_cert

    @staticmethod
    def unlock_no_ca(value: bool = True):
        """
        This function enables you to disable the CA verification of the CKAN server.

        __Warning__:
        Only allow in a local environment!
        """
        unlock_no_ca(value)

    @staticmethod
    def unlock_external_url_resource_download(value: bool = True):
        """
        This function enables the download of resources external from the CKAN server.
        """
        unlock_external_url_resource_download(value)


class DatasetParams(DatabaseParams):
    def __init__(self, source: "DatasetParams" =None):
        super().__init__(source)
        self.dataset: Union[str, None] = None
        if source is not None:
            source.copy(dest=self)

    def copy(self, *, dest=None):
        if dest is None:
            dest = DatasetParams()
        super().copy(dest=dest)
        dest.dataset = self.dataset
        return dest

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="Harvester parameters")
            DatabaseParams.setup_cli_harvester_parser(parser=parser)
        parser.add_argument("--dataset", type=str,
                            help="Dataset name")
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        super().initialize_from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        self.dataset = args.dataset


class TableParams(DatasetParams):
    def __init__(self, source: "TableParams" =None):
        super().__init__(source)
        self.output_dir: Union[str, None] = None
        self.enable_download: Union[bool, None] = None
        self.resource_url: Union[str, None] = None
        self.table: Union[str, None] = None
        self.query_string: Union[str, None] = None
        self.limit: Union[int, None] = None
        self.single_request: bool = False
        if source is not None:
            source.copy(dest=self)

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableParams()
        super().copy(dest=dest)
        dest.output_dir = self.output_dir
        dest.enable_download = self.enable_download
        dest.resource_url = self.resource_url
        dest.table = self.table
        dest.query_string = self.query_string
        dest.limit = self.limit
        dest.single_request = self.single_request
        dest.ckan_postgis = self.ckan_postgis
        return dest

    @staticmethod
    def setup_cli_harvester_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        if parser is None:
            parser = argparse.ArgumentParser(description="Harvester parameters")
            DatabaseParams.setup_cli_harvester_parser(parser)
            DatasetParams.setup_cli_harvester_parser(parser)
        parser.add_argument("-o", "--output-dir", type=str,
                            help="Output directory of download, relative to the download directory (normally provided by File/URL attribute)")  # applies to parent (builder)
        parser.add_argument("--no-download", type=bool,
                            help="Option to disable download")  # applies to parent (builder)
        parser.add_argument("--resource-url", type=str,
                            help="URL of resource")
        parser.add_argument("--table", type=str,
                            help="Table name")  # normally specified in the File/URL attribute of builder
        parser.add_argument("--query", type=str,
                            help="Query to restrict the lines of the table")
        parser.add_argument("-l", "--limit", type=int,
                            help="Number of rows per request", default=10000)
        parser.add_argument("--once",
                            help="Option to perform only one request with the default limit. This will limit the size of the Data.",
                            action="store_true", default=False)
        return parser

    def initialize_from_cli_args(self, args: argparse.Namespace, base_dir: str = None, error_not_found: bool = True,
                                 default_proxies: dict = None, proxy_headers: dict = None) -> None:
        super().initialize_from_cli_args(args, base_dir=base_dir, error_not_found=error_not_found,
                                         default_proxies=default_proxies, proxy_headers=proxy_headers)
        self.output_dir = args.output_dir  # applies to parent (builder)
        self.enable_download = not args.no_download if args.no_download is not None else None  # applies to parent (builder)
        self.resource_url = args.resource_url
        self.table = args.table
        self.limit = args.limit
        if args.once is not None:
            self.single_request = args.once

    def parse_options_string(self, options_string: str, *, base_dir: str = None, file_url_attr: str=None,
                             parser:argparse.ArgumentParser=None):
        self.file_url_attr = file_url_attr
        parser = self.setup_cli_harvester_parser(parser)
        args = parser.parse_args(shlex.split(options_string))
        self.options_string = options_string
        self.base_dir = base_dir
        self.initialize_from_cli_args(args, base_dir=base_dir)





#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a mongo database using pymongo
"""
from typing import Union, List, Any, Dict, Set
from types import SimpleNamespace
from collections import OrderedDict
from warnings import warn
import copy

import pandas as pd


try:
    import bson
except ImportError:
    bson = SimpleNamespace(ObjectId=None, DBRef=None)


from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.auxiliary.list_records import ListRecords, records_to_df
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_errors import CleanerRequirementError
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload import CkanDataCleanerUpload, _pd_series_type_detect


mongodb_keep_id_column_trace:bool = True
mongodb_keep_class_column_trace:bool = True
mongodb_id_column:str = "_id"
mongodb_id_new_column:str = "ObjectId"
mongodb_id_datatype_numeric:bool = False  # option to store ids as a numeric datatype - there is no numeric datatype which corresponds
mongodb_id_alt_type_numeric:str = "int12"  # the ObjectIds are 12 byte integers (int96) - there is no such data type in Postgre (oid?)
mongodb_dbref_extract_new_id_column_max_level:int = 1  # option to create a new column if a DBRef is found in a json field
mongodb_dbref_alt_type:str = "json"  # used if not resumed in one column


def pymongo_default_df_conversion(documents: List[dict], **kwargs) -> Union[pd.DataFrame, ListRecords]:
    df = ListRecords(documents)
    # if df.columns is not None:
    #     for i, field_name in enumerate(df.columns):
    #         if field_name == mongodb_id_column:
    #             df.columns[i] = mongodb_id_new_column
    # df = records_to_df(documents)
    return df


class BrokenMongoRefError(Exception):
    pass


class MongoDataCleanerUpload(CkanDataCleanerUpload):
    """
    Data cleaner operations specific to MongoDB objects.
    """
    def __init__(self):
        super().__init__()
        # change default parameters
        self.param_field_subs[mongodb_id_column] = mongodb_id_new_column
        self.param_field_primary_key = [mongodb_id_new_column]
        self.param_apply_field_changes = True
        # specific options
        self.param_mongodb_dbref_as_one_column:bool = True  # option to extract only the ObjectId referenced by the DBRefs
        self.collection_refs:Dict[str,Set[str]] = {}
        self.database_refs:Dict[str,Set[str]] = {}
        self.broken_collection_refs:List[str] = []
        self.broken_database_refs:List[str] = []
        if bson.DBRef is None:
            raise CleanerRequirementError("bson", "DBRef, ObjectId")

    def clear_outputs_new_dataframe(self):
        super().clear_outputs_new_dataframe()
        self.broken_collection_refs:List[str] = []
        self.broken_database_refs:List[str] = []

    def clear_all_outputs(self):
        super().clear_all_outputs()
        self.collection_refs:Dict[str,Set[str]] = {}
        self.database_refs:Dict[str,Set[str]] = {}

    def copy(self, dest=None) -> "MongoDataCleanerUpload":
        if dest is None:
            dest = MongoDataCleanerUpload()
        super().copy(dest=dest)
        dest.param_mongodb_dbref_as_one_column = self.param_mongodb_dbref_as_one_column
        dest.collection_refs = copy.deepcopy(self.collection_refs)
        dest.database_refs = copy.deepcopy(self.database_refs)
        return dest

    def _detect_standard_field_bypass(self, field_name: str, values: Union[Any, pd.Series]) -> Union[CkanField,None]:
        if _pd_series_type_detect(values, bson.DBRef):
            if self.param_mongodb_dbref_as_one_column:
                return CkanField(field_name, mongodb_id_alt_type_numeric if mongodb_id_datatype_numeric else "text")
            else:
                return CkanField(field_name, mongodb_dbref_alt_type)
        elif _pd_series_type_detect(values, bson.ObjectId):
            return CkanField(field_name, mongodb_id_alt_type_numeric if mongodb_id_datatype_numeric else "text")
        return None

    def _replace_non_standard_subvalue(self, subvalue:Any, field:CkanField, path:str, level:int,
                                       *, field_data_type:str) -> Any:
        field_name = field.name if field is not None else None
        if isinstance(subvalue, bson.ObjectId):
            if mongodb_id_datatype_numeric:
                new_subvalue = int(str(subvalue), 16)
                # new_subvalue = str(subvalue)
            else:
                new_subvalue = str(subvalue)
            if level == 0:
                self.field_suggested_index.add(field_name)
            return new_subvalue
        elif isinstance(subvalue, bson.DBRef):
            id_field = path.replace(".", "_")
            if level == 0 and self.param_mongodb_dbref_as_one_column:
                id_path = None
            elif self.param_mongodb_dbref_as_one_column:  # and level > 0
                id_path = path
            elif level == 0 and not self.param_mongodb_dbref_as_one_column:
                id_path = path + "." + mongodb_id_new_column
                id_field = id_field + "_id"
            else:
                id_path = path + "." + mongodb_id_new_column
            if (level <= mongodb_dbref_extract_new_id_column_max_level and
                    id_path is not None and id_path not in self.field_subs_path.keys()):
                self._add_field_from_path(id_path,
                                          data_type=mongodb_id_alt_type_numeric if mongodb_id_datatype_numeric else "text",
                                          new_field_name=id_field,
                                          notes=f"Column extracted from {id_path}")
            if mongodb_id_datatype_numeric:
                id_value = int(str(subvalue.id), 16)
                # id_value = str(subvalue.id)
            else:
                id_value = str(subvalue.id)
            if self.param_mongodb_dbref_as_one_column:
                if path in self.collection_refs.keys():
                    self.collection_refs[path].add(str(subvalue.collection))
                    self.database_refs[path].add(str(subvalue.database))
                else:
                    self.collection_refs[path] = {str(subvalue.collection)}
                    self.database_refs[path] = {str(subvalue.database)}
                new_subvalue = id_value
            else:
                new_subvalue = {mongodb_id_new_column: id_value,
                                "collection": subvalue.collection,
                                "database": subvalue.database,
                                }
                if id_path in self.field_subs_path.keys():
                    self._new_columns_in_row[id_path] = new_subvalue[mongodb_id_new_column]
            return new_subvalue
        elif level == 0:
            return super()._replace_non_standard_value(subvalue, field, field_data_type=field_data_type)
        else:
            return super()._replace_non_standard_subvalue(subvalue, field, path, level, field_data_type=field_data_type)

    def _replace_non_standard_value(self, value:Any, field:CkanField,
                                       *, field_data_type:str) -> Any:
        field_name = field.name if field is not None else None
        return self._replace_non_standard_subvalue(value, field, path=field_name,
                                                   level=0, field_data_type=field_data_type)

    def _extra_checks(self, records: Union[List[dict], pd.DataFrame], fields:Union[OrderedDict[str, CkanField], None]) -> None:
        self.broken_collection_refs = [path for path, refs in self.collection_refs.items() if len(refs) > 1]
        self.broken_database_refs = [path for path, refs in self.database_refs.items() if len(refs) > 1]
        if len(self.broken_collection_refs) > 0 or len(self.broken_database_refs) > 0:
            broken_refs = set(self.broken_collection_refs).union(set(self.broken_database_refs))
            msg = f"DBRefs do not point to an unique collection: {', '.join(broken_refs)}"
            if self.param_raise_error or self.param_mongodb_dbref_as_one_column:
                raise BrokenMongoRefError(msg)
            else:
                warn(msg)

def pymongo_default_data_cleaner() -> MongoDataCleanerUpload:
    return MongoDataCleanerUpload()




#!python3
# -*- coding: utf-8 -*-
"""
Section of the package dedicated to the conversion of records to a CKAN-compatible format.
This is linked to the data harvesters.
"""

from . import data_cleaner_errors
from . import data_cleaner_abc
from . import data_cleaner_upload_1_basic
from . import data_cleaner_upload_2_geom
from . import data_cleaner_upload

# usage shortcuts
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload import CkanDataCleanerUpload





#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List
import copy
from warnings import warn

from ckanapi_harvesters.auxiliary.ckan_model import CkanPackageInfo, CkanResourceInfo, CkanViewInfo
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType, assert_or_raise
from ckanapi_harvesters.auxiliary.ckan_action import CkanNotFoundError
from ckanapi_harvesters.auxiliary.ckan_errors import ReadOnlyError
from ckanapi_harvesters.ckan_api.ckan_api_1_map import use_ckan_owner_org_as_default

from ckanapi_harvesters.ckan_api.ckan_api_5_manage import CkanApiManage



class CkanApiDeprecated(CkanApiManage):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements API calls which are not recommended to use.
    """

    def copy(self, new_identifier: str = None, *, dest=None):
        if dest is None:
            dest = CkanApiDeprecated()
        super().copy(new_identifier=new_identifier, dest=dest)
        return dest

    ## Not recommended mapping functions ------------------v
    def _api_package_list(self, *, params:dict=None, owner_org:str=None, limit:int=None, offset:int=None) -> List[str]:
        """
        __Not recommended__
        API call to package_list.
        :param params: typically, the request can be limited to an organization with the owner_org parameter
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than package_list because this API does not list private packages")
        warn(msg)
        if params is None: params = {}
        if limit is None: limit = self.params.default_limit_list
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if owner_org is None and use_ckan_owner_org_as_default:
            owner_org = self.owner_org
        if owner_org is not None:
            params["owner_org"] = owner_org
        response = self._api_action_request("package_list", method=RequestType.Get, params=params)
        if response.success:
            return response.result
        else:
            raise response.default_error(self)

    def _api_package_list_all(self, *, params:dict=None, owner_org:str=None, limit:int=None, offset:int=None) -> List[str]:
        """
        __Not recommended__
        API call to package_list until an empty list is received.
        :see: api_package_list()
        :param params:
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than package_list because this API does not list private packages")
        warn(msg)
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_package_list, params=params, owner_org=owner_org, limit=limit, offset=offset)
        return sum(responses, [])

    package_list_all = _api_package_list_all  # function alias


    def _api_resource_search(self, query:str=None, *, order_by:str=None, limit:int=None, offset:int=None,
                             resource_name:str=None,
                             datastore_info:bool=None, resource_view_list:bool=None,
                             params:dict=None) -> List[CkanResourceInfo]:
        """
        __Not recommended__
        API call to resource_search. It is more recommended to use the package_show API because it is not possible to
        filter the resources by package name here. Moreover, it does not return information on private resources.
        :see: map_resources()
        :param query: (string or list of strings of the form {field}:{term1}) – The search criteria. See above for description.
        :param order_by: A field on the Resource model that orders the results.
        :param limit:
        :param offset:
        :param resource_name: a shortcut to add the filter "name:{resource_name}"
        :param datastore_info: an option to query the datastore info for all the resources found.
        If not provided, the last value for this option used with map_resources will be used.
        :param resource_view_list: an option to query the resource views list for all the resources found.
        If not provided, the last value for this option used with map_resources will be used.
        :param params: additional parameters to pass to resource_search
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than resource_search because resource_search cannot filter per package")
        warn(msg)
        if datastore_info is None:
            datastore_info = self.map._mapping_query_datastore_info
        if resource_view_list is None:
            resource_view_list = self.map._mapping_query_resource_view_list
        if params is None: params = {}
        if limit is None: limit = self.params.default_limit_list
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if query is None:
            query = []
        elif isinstance(query, str):
            query = [query]
        if resource_name is not None:
            query.append(f"name:{resource_name}")
        if query is not None:
            params["query"] = query
        if order_by is not None:
            params["order_by"] = order_by
        response = self._api_action_request("resource_search", method=RequestType.Get, params=params)
        if response.success:
            resource_info_list = [CkanResourceInfo(e) for e in response.result["results"]]
            for resource_info in resource_info_list:
                self._enrich_resource_info(resource_info, datastore_info=datastore_info,
                                           resource_view_list=resource_view_list)
            self.map._update_resource_info(resource_info_list)
            return copy.deepcopy(resource_info_list)
        else:
            raise response.default_error(self)

    def _api_resource_search_all(self, query: str = None, *, order_by: str = None, limit: int = None, offset: int = None,
                                 resource_name: str = None,
                                 datastore_info: bool = None, resource_view_list: bool = None,
                                 params: dict = None) -> List[CkanResourceInfo]:
        """
        __Not recommended__
        API call to resource_search until an empty list is received. It is more recommended to use the package_show API because it is not possible to
        filter the resources by package name here. Moreover, it does not return information on private resources.
        :see: map_resources()
        :see: _api_resource_search()
        :param query: (string or list of strings of the form {field}:{term1}) – The search criteria. See above for description.
        :param order_by: A field on the Resource model that orders the results.
        :param limit: maximum number of results to return.
        :param offset: the offset in the complete result for where the set of returned datasets should begin.
        :param resource_name: a shortcut to add the filter "name:{resource_name}"
        :param datastore_info: an option to query the datastore info for all the resources found.
        If not provided, the last value for this option used with map_resources will be used.
        :param resource_view_list: an option to query the resource views list for all the resources found.
        If not provided, the last value for this option used with map_resources will be used.
        :param params: additional parameters to pass to resource_search
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than resource_search because resource_search cannot filter per package")
        warn(msg)
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_resource_search, params=params, limit=limit, offset=offset,
                                                   query=query, order_by=order_by,
                                                   resource_name=resource_name,
                                                   datastore_info=datastore_info, resource_view_list=resource_view_list)
        return sum(responses, [])

    resource_search_all = _api_resource_search_all  # function alias


    def _api_group_package_show(self, group_name: str, *, params:dict=None, owner_org:str=None,
                                include_private:bool=True, include_drafts:bool=False, sort:str=None,
                                limit:int=None, offset:int=None) -> List[CkanPackageInfo]:
        """
        __Not recommended__
        API call to group_package_show. Return the datasets (packages) of a group.
        :param group_name: group name or id
        :param owner_org: ability to filter packages by owner_org
        :param include_private: if True, private datasets will be included in the results. Only private datasets from the user’s organizations will be returned and sysadmins will be returned all private datasets. Optional, the default is False in the API
        :param include_drafts:  if True, draft datasets will be included in the results. A user will only be returned their own draft datasets, and a sysadmin will be returned all draft datasets. Optional, the default is False.
        :param sort: sorting of the search results. Optional. Default: 'score desc, metadata_modified desc'. As per the solr documentation, this is a comma-separated string of field names and sort-orderings.
        :param limit: maximum number of results to return. Translatees to the API rows argument.
        :param offset: the offset in the complete result for where the set of returned datasets should begin. Translatees to the API start argument.
        :param params: other parameters to pass to package_search
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than group_package_show knowing the name of the package because this API does not list private packages")
        warn(msg)
        if params is None: params = {}
        params["id"] = group_name
        if limit is None: limit = self.params.default_limit_list
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if owner_org is None and use_ckan_owner_org_as_default:
            owner_org = self.owner_org
        if owner_org is not None:
            owner_org_info = self.get_organization_info_or_request(owner_org)
            owner_org = owner_org_info.id
            params["owner_org"] = owner_org
        if sort is not None:
            params["sort"] = sort
        if include_private is not None:
            params["include_private"] = include_private
        if include_drafts is not None:
            params["include_drafts"] = include_drafts
        response = self._api_action_request("group_package_show", method=RequestType.Get, params=params)
        if response.success:
            package_info_list = [CkanPackageInfo(e) for e in response.result]
            self.map._update_package_info(package_info_list)
            return package_info_list
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "Group", response)
        else:
            raise response.default_error(self)

    def _api_group_package_show_all(self, group_name: str, *, params:dict=None, owner_org:str=None,
                                    include_private:bool=True, include_drafts:bool=False, sort:str=None,
                                    limit:int=None, offset:int=None) -> List[CkanPackageInfo]:
        """
        __Not recommended__
        API call to group_package_show until an empty list is received.
        :see: _api_group_package_show()
        :param group_name: group name or id
        :param owner_org: ability to filter packages by owner_org
        :param include_private: if True, private datasets will be included in the results. Only private datasets from the user’s organizations will be returned and sysadmins will be returned all private datasets. Optional, the default is False in the API
        :param include_drafts:  if True, draft datasets will be included in the results. A user will only be returned their own draft datasets, and a sysadmin will be returned all draft datasets. Optional, the default is False.
        :param sort: sorting of the search results. Optional. Default: 'score desc, metadata_modified desc'. As per the solr documentation, this is a comma-separated string of field names and sort-orderings.
        :param limit: maximum number of results to return. Translatees to the API rows argument.
        :param offset: the offset in the complete result for where the set of returned datasets should begin. Translatees to the API start argument.
        :param params: other parameters to pass to API
        :return:
        """
        msg = DeprecationWarning("Prefer using package_search rather than group_package_show knowing the name of the package because this API does not list private packages")
        warn(msg)
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_group_package_show, params=params, limit=limit, offset=offset,
                                                   group_name=group_name, owner_org=owner_org,
                                                   include_private=include_private, include_drafts=include_drafts)
        return sum(responses, [])

    group_package_show_all = _api_group_package_show_all  # function alias

    ## resource view
    def _api_resource_create_default_resource_views(self, resource_id:str, *, create_datastore_views:bool=None,
                                                    params:dict=None) -> List[CkanViewInfo]:
        """
        API call to resource_create_default_resource_views
        :param resource_id: resource id
        :param create_datastore_views: whether to create views that rely on data being on the DataStore (optional, API defaults to False)
        :param params:
        :return:
        """
        msg = DeprecationWarning("Prefer using resource_view_create rather than resource_create_default_resource_views")
        warn(msg)
        assert_or_raise(not self.params.read_only, ReadOnlyError())
        if params is None: params = {}
        resource_info = self.resource_show(resource_id)
        resource_dict = resource_info.details
        if create_datastore_views is None:
            create_datastore_views = self.resource_is_datastore(resource_id)
        params["resource"] = resource_dict
        params["create_datastore_views"] = create_datastore_views
        response = self._api_action_request(f"resource_create_default_resource_views", method=RequestType.Post, json=params)
        if response.success:
            view_info_list = [CkanViewInfo(view_dict) for view_dict in response.result]
            self.map._update_view_info(view_info_list)
            return copy.deepcopy(view_info_list)
        else:
            raise response.default_error(self)

    resource_create_default_resource_views = _api_resource_create_default_resource_views  # function alias





#!python3
# -*- coding: utf-8 -*-
"""
Package with helper function for CKAN requests using pandas DataFrames.
"""

from . import ckan_defs
from . import path
from . import login
from . import urls
from . import proxy_config
from . import external_code_import
from . import list_records
from . import ckan_action
from . import ckan_errors
from . import ckan_configuration
from . import ckan_api_key
from . import ckan_model
from . import ckan_map
from . import ckan_vocabulary_deprecated
from . import ckan_auxiliary
from . import deprecated

from .ckan_map import CkanMap
from .external_code_import import unlock_external_code_execution




#!python3
# -*- coding: utf-8 -*-
"""
Section of the package dedicated to the harvesting of data using APIs, or databases
"""

from . import file_formats
from . import data_cleaner

from . import harvester_errors
from . import harvester_model
from . import harvester_params
from . import harvester_abc
from . import pymongo_data_cleaner
from . import pymongo_params
from . import pymongo_harvester
from . import postgre_params
from . import postgre_harvester
from . import harvester_init

# usage shortcuts





#!python3
# -*- coding: utf-8 -*-
"""
Setting the proxy from simple command line arguments
"""
import urllib.request
from typing import Union, Sequence, Tuple
import os
import argparse
from warnings import warn
import copy
import json

from ckanapi_harvesters.auxiliary.ckan_defs import environ_keyword
from ckanapi_harvesters.auxiliary.path import sanitize_path, path_rel_to_dir
from ckanapi_harvesters.auxiliary.login import Login

import requests
from requests.auth import AuthBase, HTTPProxyAuth, HTTPBasicAuth

PROXY_AUTH_ENVIRON = "PROXY_AUTH_FILE"

class HttpsProxyDefError(Exception):
    def __init__(self):
        super().__init__("Only one of http_proxy or https_proxy is set")

def get_proxies_from_environ() -> dict:
    proxies = urllib.request.getproxies()
    return proxies
    # http_proxy = os.environ.get("http_proxy")
    # https_proxy = os.environ.get("https_proxy")
    # no_proxy = os.environ.get("no_proxy")
    # if http_proxy is not None and https_proxy is not None:
    #     proxies = {"http": http_proxy, "https": https_proxy}
    # elif http_proxy is not None:
    #     proxies = {"http": http_proxy, "https": http_proxy}
    # elif https_proxy is not None:
    #     raise HttpsProxyDefError()
    # else:
    #     proxies = None
    # if proxies is not None and no_proxy is not None:
    #     proxies["no"] = no_proxy
    # return proxies

def host_port_sep(url:Union[str,None], *, default_port:int=None) -> Tuple[Union[str,None],Union[int,None]]:
    if url is None:
        return None, None
    if ':' in url:
        host_prefix, host_suffix = "", ""
        if '@' in url:
            host_prefix, url = url.split('@')
        if '?' in url:
            url, host_suffix = url.split('?')
        host, port_str = url.split(':')
        host = host_prefix + host + host_suffix
        port = int(port_str)
    else:
        host, port = url, default_port
    return host, port

def _define_proxies(proxy_string:Union[str, dict], default_proxies:dict=None) -> dict:
    if proxy_string is None:
        proxies = None
    elif isinstance(proxy_string, dict):
        proxies = proxy_string
    elif isinstance(proxy_string, str):
        proxy_string = proxy_string.strip()
        proxy_mode = proxy_string.lower()
        if proxy_mode == environ_keyword:
            proxies = get_proxies_from_environ()
            if proxies is None:
                proxies = default_proxies
        elif proxy_mode == "unspecified":
            proxies = None  # do not specify the proxies - is equivalent to "environ"
        elif proxy_mode == "noproxy":
            proxies = {"http": "", "https": ""}  # do not use any proxy
        elif proxy_mode == "default":
            proxies = default_proxies  # default proxies, provided in argument
        elif proxy_string.startswith('{'):
            # proxy string is a string representation of proxy dictionary
            proxies = json.loads(proxy_string)
        else:
            # suppose string contains an url to a proxy server
            proxies = {"http": proxy_string, "https": proxy_string}
            # if "http" not in proxy_string:
            #     # url without http
            #     proxies = {"http": f"http://{proxy_string}", "https": f"http://{proxy_string}"}
            # else:
            #     proxies = {"http": proxy_string, "https": proxy_string}
    else:
        raise TypeError("proxy must be str or dict")
    return proxies


class ProxyConfig:
    def __init__(self, proxy_string:Union[str,dict]=None, default_proxies:dict=None,
                 proxy_headers:dict = None, proxy_auth:Union[AuthBase, Tuple[str,str]]=None) -> None:
        """
        :param proxy_string: string or proxies dict or ProxyConfig object.
        If a string is provided, it must be an url to a proxy or one of the following values:
            - "environ": use the proxies specified in the environment variables "http_proxy" and "https_proxy"
            - "noproxy": do not use any proxies
            - "unspecified": do not specify the proxies
            - "default": use value provided by default_proxies
        :param default_proxies: proxies used if proxies="default"
        :param proxy_headers: headers used to access the proxies, generally for authentication
        """
        if proxy_headers is None: proxy_headers = {}
        self._proxy_string:Union[str, dict, None] = None
        self._proxies:Union[dict,None] = None
        self._is_defined:bool = False
        self._default_proxies:Union[dict,None] = default_proxies
        self.proxy_headers: dict = proxy_headers
        self._proxy_auth: Union[AuthBase, Tuple[str,str], None] = proxy_auth
        self.proxy_auth_file: Union[str,None] = None
        self.proxy_auth_from_env: bool = False
        self.proxy_string = proxy_string  # property
        # self.load_proxy_auth_environ(error_not_found=False)  # recommended to base these parameters on user demand (confirm if there is a risk of leakage)

    def __str__(self):
        return str(self._proxies)

    def __copy__(self):
        return self.copy()

    def copy(self) -> "ProxyConfig":
        dest = ProxyConfig()
        dest._proxies = copy.deepcopy(self._proxies)
        dest._proxy_string = copy.deepcopy(self._proxy_string)
        dest._default_proxies = copy.deepcopy(self._default_proxies)
        dest.proxy_headers = copy.deepcopy(self.proxy_headers)
        dest._proxy_auth = copy.deepcopy(self._proxy_auth)
        dest._is_defined = self._is_defined
        dest.proxy_auth_file = self.proxy_auth_file
        dest.proxy_auth_from_env = self.proxy_auth_from_env
        return dest


    @property
    def proxy_string(self) -> Union[str, dict, None]:
        return self._proxy_string
    @proxy_string.setter
    def proxy_string(self, proxy_string:Union[str, dict, None]):
        self._proxy_string = proxy_string
        self._is_defined = proxy_string is not None
        self._proxies = _define_proxies(proxy_string, default_proxies=self._default_proxies)
    @property
    def proxies(self) -> dict:
        return self._proxies
    @proxies.setter
    def proxies(self, proxies:dict):
        self._proxy_string = proxies
        self._proxies = proxies
        self._is_defined = True
    @property
    def proxy_auth(self) -> Union[AuthBase, Tuple[str,str]]:
        return self._proxy_auth
    @proxy_auth.setter
    def proxy_auth(self, proxy_auth:Union[AuthBase, Tuple[str,str]]):
        self._proxy_auth = proxy_auth
        self.proxy_auth_file = None
        self.proxy_auth_from_env = False

    def get_host_port(self) -> Tuple[Union[str,None],Union[int,None]]:
        # special mode
        if self._proxies is None:
            return None, None
        elif "http_proxy" in self._proxies.keys() and self._proxies["http_proxy"] is not None:
            return host_port_sep(self._proxies["http_proxy"])
        else:
            return None, None

    def get_proxy_login(self) -> Login:
        if self._proxy_auth is None:
            return Login()
        else:
            assert(isinstance(self._proxy_auth, HTTPBasicAuth))  # HTTPProxyAuth is a super class of HTTPBasicAuth
            return Login(self._proxy_auth.username, self._proxy_auth.password)

    @staticmethod
    def from_str_or_config(proxies:Union[str,dict, "ProxyConfig"],
                           *, default_proxies:dict=None, proxy_headers:dict=None) -> "ProxyConfig":
        if proxies is None:
            return ProxyConfig(None, default_proxies=default_proxies , proxy_headers=proxy_headers)
        elif isinstance(proxies, ProxyConfig):
            if proxy_headers is not None:
                proxies.proxy_headers = proxy_headers
            return proxies
        else:
            return ProxyConfig(proxies, default_proxies=default_proxies, proxy_headers=proxy_headers)

    def replace_default_proxy(self, default_proxies:dict) -> None:
        if self._proxy_string is not None and self._proxy_string.lower() == "default":
            self._proxies = default_proxies

    def reset(self) -> None:
        self._proxy_string = None
        self._proxies = None
        self._is_defined = False

    def is_defined(self) -> bool:
        return self._is_defined

    def load_proxy_auth_environ(self, *, error_not_found:bool=False) -> bool:
        proxy_auth_file = sanitize_path(os.environ.get(PROXY_AUTH_ENVIRON))  # "PROXY_AUTH_FILE"
        if proxy_auth_file is not None:
            proxy_keyword = proxy_auth_file.strip().lower()
            assert(not proxy_keyword == environ_keyword)  # this value would create an infinite loop
            if self.load_proxy_auth_from_file(proxy_auth_file, error_not_found=error_not_found):
                self.proxy_auth_from_env = True
                return True
        return False

    def load_proxy_auth_from_file(self, file_path:str, *, base_dir:str=None, error_not_found:bool=True) -> bool:
        file_path = path_rel_to_dir(file_path, base_dir=base_dir, keyword_exceptions={environ_keyword})
        proxy_keyword = file_path.strip().lower()
        if proxy_keyword == environ_keyword:
            # this keyword is not very useful if proxy authentication file is loaded from environment anyway
            return self.load_proxy_auth_environ(error_not_found=error_not_found)
        if (not error_not_found) and (not os.path.isfile(file_path)):
            msg = f"Proxy authentication file does not exist: {file_path}"
            warn(msg)
            return False
        self.proxy_auth_file = file_path
        self.proxy_auth_from_env = False
        with open(file_path, "r") as f:
            auth_type = f.readline().strip().lower()
            username = f.readline().strip()
            password = f.readline().strip()
            if auth_type == "basic" or auth_type == "httpbasicauth":
                self._proxy_auth = requests.auth.HTTPBasicAuth(username, password)
            elif auth_type == "proxy" or auth_type == "httpproxyauth":
                self._proxy_auth = requests.auth.HTTPProxyAuth(username, password)
            elif auth_type == "digest" or auth_type == "httpdigestauth":
                self._proxy_auth = requests.auth.HTTPDigestAuth(username, password)
            elif auth_type == "none":
                self._proxy_auth = None
            else:
                raise KeyError(f"Unknown auth type {auth_type}")
        return True


    @staticmethod
    def _setup_cli_proxy_parser(parser: argparse.ArgumentParser = None) -> argparse.ArgumentParser:
        """
        Define or add CLI arguments to initialize the proxy
        parser help message:

        Proxy parameters initialization

        options:
          -h, --help            show this help message and exit
          --proxy PROXY         Proxy for HTTP and HTTPS

        :param parser: option to provide an existing parser to add the specific fields needed to initialize a CKAN API connection
        :return:
        """
        if parser is None:
            parser = argparse.ArgumentParser(description="Proxy parameters initialization")
        parser.add_argument("--proxy", type=str,
                            help="Proxy for HTTP and HTTPS")
        parser.add_argument("--http-proxy", type=str,
                            help="HTTP proxy")
        parser.add_argument("--https-proxy", type=str,
                            help="HTTPS proxy")
        parser.add_argument("--no-proxy", type=str,
                            help="Proxy exceptions")
        parser.add_argument("--proxy-auth-file", type=str,
                            help="Path to a proxy authentication file with 3 lines (authentication method, username, password)")
        return parser

    @staticmethod
    def from_cli_args(args: argparse.Namespace, *, base_dir:str=None, error_not_found:bool=True,
                      default_proxies:dict=None, proxy_headers:dict=None) -> "ProxyConfig":
        proxy_string, proxies = None, None
        if args.proxy is not None:
            proxy_string = args.proxy
        elif args.http_proxy is not None and args.https_proxy is not None:
            proxies = {"http": args.http_proxy, "https": args.https_proxy}
        elif args.http_proxy is not None:
            proxies = {"http": args.http_proxy, "https": args.http_proxy}
        if proxies is not None and args.no_proxy is not None:
            proxies["no"] = args.no_proxy
        elif args.https_proxy is not None:
            raise HttpsProxyDefError()
        if proxy_string is not None:
            proxy_config = ProxyConfig(proxy_string, default_proxies=default_proxies, proxy_headers=proxy_headers)
        elif proxies is not None:
            proxy_config = ProxyConfig(proxies, default_proxies=default_proxies, proxy_headers=proxy_headers)
        else:
            proxy_config = None
        if args.proxy_auth_file is not None:
            if proxy_config is not None:
                proxy_config.load_proxy_auth_from_file(args.proxy_auth_file, base_dir=base_dir, error_not_found=error_not_found)
            else:
                raise Exception(f"Proxy authentication file specified without proxy specification: {args.proxy_auth_file}")
        return proxy_config




#!python3
# -*- coding: utf-8 -*-
"""
Detailed report on package resources: size, access rights and data format policy scores
"""
from typing import List, Union, Dict
from collections import OrderedDict
import time
import datetime
import os
from warnings import warn

from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import to_jsons_indent_lists_single_line
from ckanapi_harvesters.auxiliary.ckan_model import CkanVisibility, CkanUserInfo
from ckanapi_harvesters.policies.data_format_policy_errors import ErrorCount, DataPolicyError


def round_size(value_mb:float) -> float:
    return round(value_mb, 2)


class CkanAdminReport:
    def __init__(self, package_list:List[str]=None, cancel_if_present:bool=True,
                 package_custom_fields:List[str]=None, ckan:CkanApi=None, full_report:bool=False):
        if package_custom_fields is None:
            package_custom_fields = []  # option to include specific custom fields in the report e.g. a end of license date
        if isinstance(package_list, str):
            package_list = [package_list]
        self.package_list: Union[List[str],None] = package_list
        self.cancel_if_present: bool = cancel_if_present
        self.include_package_custom_fields: List[str] = package_custom_fields
        self.include_resources_detail: bool = True
        self.include_policy_messages: bool = full_report
        self.include_group_report: bool = full_report
        self.date_format:Union[str,None] = '%d/%m/%Y %H:%M'
        self._connected_user: Union[CkanUserInfo, None] = None
        self.report_date: Union[datetime.datetime, None] = None
        self._elapsed_time_requests: Union[float,None] = None
        self._request_count: Union[int,None] = None
        self.report: Union[dict,None] = None  # report output
        if ckan is not None:
            self.execute(ckan)

    def _date_format_str(self, date:datetime.datetime) -> str:
        if self.date_format is not None:
            return date.strftime(self.date_format)
        else:
            return date.isoformat()

    def _perform_requests(self, ckan: CkanApi) -> None:
        if not self.cancel_if_present:
            ckan.purge(purge_map=True)
        start = time.time()
        self.report_date = datetime.datetime.now()
        request_count_init = ckan.debug.ckan_request_counter
        self._connected_user = ckan.query_current_user()
        if not self._connected_user.sysadmin:
            msg = f"It is recommended to run the report with a user with sysadmin rights. Current user: {self._connected_user.name}"
            warn(msg)
        ckan.map_resources(self.package_list, datastore_info=True, only_missing=self.cancel_if_present)
        ckan.organization_list_all(cancel_if_present=False, include_users=True)
        ckan.license_list(cancel_if_present=self.cancel_if_present)
        ckan.map_file_resource_sizes(cancel_if_present=self.cancel_if_present)
        ckan.map_user_rights(cancel_if_present=self.cancel_if_present)
        self._elapsed_time_requests = time.time() - start
        self._request_count = ckan.debug.ckan_request_counter - request_count_init

    def _consolidate(self, ckan: CkanApi) -> None:
        for user_info in ckan.map.users.values():
            user_info.organizations = []
        for organization_info in ckan.map.organizations.values():
            for user_id in organization_info.user_members.keys():
                ckan.map.users[user_id].organizations.append(organization_info.name)

    def _create_report(self, ckan: CkanApi) -> None:
        policy_messages: Dict[str, List[DataPolicyError]] = {}
        ckan.policy_check(buffer=policy_messages)

        report_header = OrderedDict([
            ("title", "Admin report on packages and resources"),
            ("date", self._date_format_str(self.report_date)),
            ("ckan", ckan.url),
            ("user", self._connected_user.name),
            ("user_sysadmin", self._connected_user.sysadmin),
            ("package_selection", self.package_list if self.package_list is not None else "All"),
        ])
        packages_report = {}
        total_policy_errors = ErrorCount([])
        total_filestore_size_mb = 0.
        total_external_size_mb = 0.
        total_datastore_size_mb = 0.
        total_resource_count = 0
        total_external_resource_count = 0
        total_datastore_count = 0
        total_datastore_lines = 0
        global_last_modified_resources = None
        global_last_modified_metadata = None
        for package_id, package_info in ckan.map.packages.items():
            package_name = package_info.name
            package_data_format_messages = policy_messages.get(package_name, [])
            data_format_policy_scores = ErrorCount(package_data_format_messages)
            total_policy_errors += data_format_policy_scores
            resources_report = []
            last_modified_resource = None
            last_modified_resource_metadata = None
            package_resource_count = len(package_info.package_resources)
            package_external_resource_count = 0
            package_datastore_count = 0
            package_filestore_size_mb = 0.
            package_external_size_mb = 0.
            package_datastore_size_mb = 0.
            package_datastore_lines = 0
            for resource_id, resource_info in package_info.package_resources.items():
                resource_modified = resource_info.last_modified if resource_info.last_modified is not None else resource_info.created
                internal_filestore = ckan.is_url_internal(resource_info.download_url)
                resource_report = OrderedDict([
                    ("resource_name", resource_info.name),
                    ("id", resource_id),
                    ("state", str(resource_info.state)),
                    ("external_url", resource_info.download_url if resource_info.download_url and not internal_filestore else None),
                    ("filestore_size_mb", resource_info.download_size_mb if internal_filestore else None),
                    ("external_size_mb", resource_info.download_size_mb if not internal_filestore else None),
                    ("datastore_size_mb", 0),
                    ("datastore_active", resource_info.datastore_active),
                    ("datastore_lines", None),
                    ("date_modified", self._date_format_str(resource_modified) if resource_modified is not None else None),
                    ("metadata_modified", self._date_format_str(resource_info.metadata_modified) if resource_info.metadata_modified is not None else None),
                    ("datastore_aliases", None),
                ])
                if resource_modified is not None:
                    last_modified_resource = max(last_modified_resource, resource_modified) \
                        if last_modified_resource else resource_modified
                    global_last_modified_resources = max(global_last_modified_resources, resource_modified) \
                        if global_last_modified_resources else resource_modified
                if resource_info.metadata_modified is not None:
                    last_modified_resource_metadata = max(last_modified_resource_metadata, resource_info.metadata_modified) \
                        if last_modified_resource_metadata else resource_info.metadata_modified
                    global_last_modified_metadata = max(global_last_modified_metadata, resource_modified) \
                        if global_last_modified_metadata else resource_modified
                if resource_info.download_url:
                    if internal_filestore:
                        package_filestore_size_mb += resource_info.download_size_mb
                    else:
                        package_external_size_mb += resource_info.download_size_mb
                        package_external_resource_count += 1
                if resource_info.datastore_info is not None:
                    datastore_size = round_size(resource_info.datastore_info.table_size_mb + resource_info.datastore_info.index_size_mb)
                    resource_report["datastore_aliases"] = resource_info.datastore_info.aliases
                    resource_report["datastore_size_mb"] = datastore_size
                    package_datastore_size_mb += datastore_size
                    resource_report["datastore_lines"] = resource_info.datastore_info.row_count
                    package_datastore_lines += resource_info.datastore_info.row_count
                    package_datastore_count += 1
                resources_report.append(resource_report)
            package_report = OrderedDict([
                ("package_title", package_info.title),
                ("state", str(package_info.state)),
                ("organization", package_info.organization_info.name if package_info.organization_info else None),
                ("version", package_info.version),
                ("license", ckan.map.licenses[package_info.license_id].title if package_info.license_id else None),
                ("license_domain", ckan.map.licenses[package_info.license_id].domain.to_dict() if package_info.license_id else None),
                ("author", package_info.author),
                ("maintainer", package_info.maintainer),
                ("metadata_modified", self._date_format_str(package_info.metadata_modified)),
                ("resources_modified", self._date_format_str(last_modified_resource) if last_modified_resource is not None else None),
                ("resources_metadata_modified", self._date_format_str(last_modified_resource_metadata) if last_modified_resource_metadata is not None else None),
                ("visibility", str(CkanVisibility.from_bool_is_private(package_info.private))),
                ("filestore_total_size_mb", round_size(package_filestore_size_mb)),
                ("external_total_size_mb", round_size(package_external_size_mb)),
                ("datastore_total_size_mb", round_size(package_datastore_size_mb)),
                ("datastore_total_lines", package_datastore_lines),
                ("resource_count", package_resource_count),
                ("among_resources_external", package_external_resource_count),
                ("among_resources_datastore", package_datastore_count),
                ("data_format_policy_scores", data_format_policy_scores.to_dict()),
                ("tags", package_info.tags),
            ])
            for custom_field in self.include_package_custom_fields:
                package_report[custom_field] = package_info.custom_fields.get(custom_field, None)
            package_report["users"] = []
            package_report["groups"] = []
            if self.include_resources_detail:
                package_report["resources"] = resources_report
            if package_info.private:
                users_dict = {ckan.map.users[user_id].name: collaboration.to_dict(ckan.map.users[user_id], ckan.map.groups, self.date_format)
                              for user_id, collaboration in package_info.user_access.items()}
                package_report["users"] = OrderedDict(sorted(users_dict.items()))
            else:
                # TODO: do all users have write access if package is Public
                package_report["users"] = "all (Public)"
            package_report["groups"] = sorted([group_info.name for group_info in package_info.groups])
            if self.include_policy_messages:
                package_report["policy_messages"] = [message.to_dict() for message in package_data_format_messages]
            total_filestore_size_mb += package_filestore_size_mb
            total_external_size_mb += package_external_size_mb
            total_datastore_size_mb += package_datastore_size_mb
            total_resource_count += package_resource_count
            total_external_resource_count += package_external_resource_count
            total_datastore_count += package_datastore_count
            total_datastore_lines += package_datastore_lines
            global_last_modified_metadata = max(global_last_modified_metadata, package_info.metadata_modified) \
                if global_last_modified_metadata else package_info.metadata_modified
            packages_report[package_name] = package_report
        packages_report = OrderedDict(sorted(packages_report.items()))
        report_totals = OrderedDict([
            ("total_filestore_size_mb", round_size(total_filestore_size_mb)),
            ("total_datastore_size_mb", round_size(total_datastore_size_mb)),
            ("total_external_size_mb", round_size(total_external_size_mb)),
            ("total_datastore_lines", total_datastore_lines),
            ("num_packages", len(packages_report)),
            ("total_resource_count", total_resource_count),
            ("among_resources_external", total_external_resource_count),
            ("among_resources_datastore", total_datastore_count),
            ("last_modified_data", self._date_format_str(global_last_modified_resources) if global_last_modified_resources else None),
            ("last_modified_metadata", self._date_format_str(global_last_modified_metadata) if global_last_modified_metadata else None),
            ("total_policy_errors", total_policy_errors.to_dict()),
        ])
        sysadmin_report = {user_info.name: OrderedDict([
            ("fullname", user_info.fullname),
            ("last_active", self._date_format_str(user_info.last_active) if user_info.last_active is not None else None),
            ("organizations", user_info.organizations),
        ]) for user_info in ckan.map.users.values() if user_info.sysadmin}
        sysadmin_report = OrderedDict(sorted(sysadmin_report.items()))
        users_report = {user_info.name: OrderedDict([
            ("fullname", user_info.fullname),
            ("last_active", self._date_format_str(user_info.last_active) if user_info.last_active is not None else None),
            ("organizations", user_info.organizations),
         ]) for user_info in ckan.map.users.values() if not user_info.sysadmin}
        users_report = OrderedDict(sorted(users_report.items()))
        groups_report = {group_info.name: OrderedDict([
            ("group_title", group_info.title),
            ("package_count", group_info.package_count),
            ("users_count", len(group_info.user_members)),
            ("users", OrderedDict(sorted({ckan.map.users[user_id].name: str(capacity) for user_id, capacity in group_info.user_members.items()}.items())) if group_info.user_members is not None else None),
        ]) for group_info in ckan.map.groups.values()}
        groups_report = OrderedDict(sorted(groups_report.items()))
        report_footer = OrderedDict([
            ("requests_count", self._request_count),
            ("time_elapsed_seconds", self._elapsed_time_requests),
        ])
        report = OrderedDict([
            ("header", report_header),
            ("totals", report_totals),
            ("packages", packages_report),
            ("users", OrderedDict([
                ("sysadmins", sysadmin_report),
                ("other", users_report),
            ])),
        ])
        if self.include_group_report:
            report["groups"] = groups_report
        report["footer"] = report_footer
        self.report = report

    def execute(self, ckan: CkanApi) -> dict:
        self._perform_requests(ckan)
        self._consolidate(ckan)
        self._create_report(ckan)
        return self.report

    def refresh_report(self, ckan: CkanApi) -> dict:
        self._create_report(ckan)
        return self.report

    def to_jsons(self) -> str:
        return to_jsons_indent_lists_single_line(self.report)

    def to_json(self, file_path:str) -> None:
        with open(file_path, "w", encoding="utf8") as f:
            f.write(self.to_jsons())


if __name__ == '__main__':
    ckan = CkanApi()
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)

    package_list = None  # use this argument or no argument to make a full report
    # package_list = ["builder-example-py"]  # limit to the example package

    report = CkanAdminReport(ckan=ckan, package_list=package_list, full_report=True)
    print(report.to_jsons())

    self_dir = os.path.abspath(os.path.dirname(os.path.realpath(__file__)))
    top_dir = os.path.abspath(os.path.join(self_dir, "..", ".."))
    tests_dir = os.path.abspath(os.path.join(top_dir, "..", "tests"))
    out_file = os.path.join(tests_dir, "admin_report.json")
    # out_file = os.path.join(tests_dir, f"admin_report_{report.report_date.strftime('%Y%m%dT%H%M')}.json")
    report.to_json(out_file)

    print(f"Done. Saved report to {out_file}")



#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore without uploading any data.
"""
import time
from abc import ABC, abstractmethod
from typing import Dict, List, Callable, Any, Tuple, Union, Set
import os
import io
from warnings import warn

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.builder.builder_resource import builder_request_default_auth_if_ckan
from ckanapi_harvesters.builder.builder_resource_datastore import BuilderDataStoreFile
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, DataStoreNotFoundError
from ckanapi_harvesters.builder.builder_errors import RequiredDataFrameFieldsError, ResourceFileNotExistMessage
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo, CkanDataStoreInfo
from ckanapi_harvesters.auxiliary.ckan_errors import CkanArgumentError, FunctionMissingArgumentError, ExternalUrlLockedError
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, assert_or_raise, find_duplicates, datastore_id_col
from ckanapi_harvesters.ckan_api.ckan_api_2_readonly import df_download_read_csv_kwargs


class BuilderDataStoreUrl(BuilderDataStoreFile):  #, BuilderUrlABC):  # multiple inheritance can give undefined results
    """
    Class representing a DataStore (resource metadata and fields metadata) defined by a url.
    """
    def __init__(self, *, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, url:str=None):
        super(BuilderDataStoreFile, self).__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url)
        # super(BuilderUrlABC, self).__init__(name=name, format=format, description=description, resource_id=resource_id, download_url=download_url, url=url)
        self.reupload_on_update = False
        self.reupload_if_needed = False
        self.url:str = url
        self.file_name = name

    def copy(self, *, dest=None):
        if dest is None:
            dest = BuilderDataStoreUrl()
        super().copy(dest=dest)
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.url = self.url
        dest.file_name = self.file_name
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super(BuilderDataStoreFile, self)._load_from_df_row(row=row)
        # super(BuilderUrlABC, self)._load_from_df_row(row=row)
        self.url: str = _string_from_element(row["file/url"])
        self.file_name = self.name

    @staticmethod
    def sample_file_path_is_url() -> bool:
        return True

    def get_sample_file_path(self, resources_base_dir: str) -> str:
        return self.url

    def load_sample_data(self, resources_base_dir:str, *, ckan:CkanApi=None,
                         proxies:dict=None, headers:dict=None) -> bytes:
        self.sample_source = self.url
        if ckan is None:
            raise FunctionMissingArgumentError("BuilderDataStoreUrl.load_sample_data", "ckan")
        return ckan.download_url_proxy(self.url, proxies=proxies, headers=headers, auth_if_ckan=builder_request_default_auth_if_ckan).content

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True) -> pd.DataFrame:
        payload = self.load_sample_data(resources_base_dir=resources_base_dir)
        buffer = io.StringIO(payload.decode())
        response_df = self.local_file_format.read_buffer(buffer, fields=self._get_fields_info())
        if upload_alter:
            df_upload = self.df_mapper.df_upload_alter(response_df, self.sample_data_source, fields=self._get_fields_info())
            return df_upload
        else:
            return response_df

    @staticmethod
    def resource_mode_str() -> str:
        return "DataStore from URL"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = self.url
        return d

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        if ckan is None:
            return ResourceFileNotExistMessage(self.name, ErrorLevel.Warning, "Could not determine if resource url exists because ckan argument was not provided.")
        else:
            return ckan.download_url_proxy_test_head(self.url, **kwargs)

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload:pd.DataFrame=None, payload:Union[bytes, io.BufferedIOBase]=None,
                      reupload: bool = None, resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Specific implementation of patch_request which does not upload any data and only updates the fields currently present in the database
        :param resources_base_dir:
        :param ckan:
        :param package_id:
        :param reupload:
        :return:
        """
        if reupload is None: reupload = self.reupload_on_update
        if payload is not None or df_upload is not None:
            raise CkanArgumentError("payload", "datastore defined from URL patch")
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        try:
            df_download = self.download_sample_df(ckan, download_alter=False, search_all=False, limit=1)
            if df_download is None:
                assert_or_raise(resource_id is None, RuntimeError("Unexpected: resource_id should be None"))
                raise NotMappedObjectNameError(self.name)
            current_fields = set(df_download.columns)
        except NotMappedObjectNameError as e:
            df_download = None
            current_fields = set()
        except DataStoreNotFoundError as e:
            df_download = None
            current_fields = set()
        empty_datastore = df_download is None or len(df_download) == 0
        data_cleaner_fields = None
        data_cleaner_index = set()
        current_fields -= {datastore_id_col}  # _id does not require documentation
        aliases = self._get_alias_list(ckan)
        self._check_necessary_fields(current_fields, raise_error=False, empty_datastore=empty_datastore)
        self._check_undocumented_fields(current_fields)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_fields=current_fields,
                                                             error_missing=False, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_fields, data_cleaner_fields, reupload=reupload)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.format, description=self.description, state=self.state,
                                             url=self.url,
                                             datastore_create=False, auto_submit=False, create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, aliases=aliases, reupload=False, data_cleaner=self.data_cleaner_upload)
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_fields, ckan)
        if reupload:
            # re-initialize datastore to reupload from url
            # normally, data was automatically submitted to DataStore on resource_create (not needed)
            ckan.datastore_create(resource_id, fields=fields, primary_key=primary_key, indexes=indexes, aliases=aliases)
            ckan.datastore_submit(resource_id)
        return resource_info



